//Arduino 1.0+ only
#include <Wire.h>
#include <E24C1024.h>

//Rocket MEMS IMU shield sketch by Dmitry Platonov. (c) 2012
//Based on code by Ben Gatti & others.

#define L3G4200D_I2C_ADDR 105
#define L3G4200D_CTRL_REG1 0x20
#define L3G4200D_CTRL_REG2 0x21
#define L3G4200D_CTRL_REG3 0x22
#define L3G4200D_CTRL_REG4 0x23
#define L3G4200D_CTRL_REG5 0x24
#define L3G4200D_X_LOW_REG  0x28
#define L3G4200D_X_HIGH_REG 0x29
#define L3G4200D_Y_LOW_REG  0x2A
#define L3G4200D_Y_HIGH_REG 0x2B
#define L3G4200D_Z_LOW_REG  0x2C
#define L3G4200D_Z_HIGH_REG 0x2D

#define LIS331DL_I2C_ADDR 0b0011101
#define LIS331DL_CTRL_REG1 0x20
#define LIS331DL_CTRL_REG2 0x21
#define LIS331DL_STATUS_REG 0x27
#define LIS331DL_X_HIGH_REG 0x29
#define LIS331DL_Y_HIGH_REG 0x2B
#define LIS331DL_Z_HIGH_REG 0x2D

#define BUTTON_PIN 4

//Must be exactly 16 bytes long
struct imu_data{
  int8_t _pad;    // First byte in page is written incorrectly to EEPROM, this is workaround
  uint32_t time;
  int16_t gyro_x;
  int16_t gyro_y;
  int16_t gyro_z;
  int8_t accel_x;
  int8_t accel_y;
  int8_t accel_z;
  uint16_t magic; // Dataset number
} __attribute__((packed)); 

void setup(){ 
  Wire.begin();
  Serial.begin(115200);
  if(sizeof(struct imu_data) != 16){
    Serial.println("sizeof(struct imu_data) != 16; SYSTEM HALTED");
    while(1);
  }

  Serial.println("Configuring L3G4200D");
  setupL3G4200D(2000); // Configure L3G4200  - 250, 500 or 2000 deg/sec

  Serial.println("Configuring LIS331D");
  setupLIS331DL();

  pinMode(BUTTON_PIN, INPUT); 

  delay(1500); //FIXME - is it needed?
}

enum mode {IDLE, RECORD, DUMP}; 

void loop(){
  struct imu_data d;
  static unsigned long lastTime = 0;
  static unsigned long pressTime = 0;
  static int buttonState = LOW;
  static int prevButtonState = LOW;
  static mode curMode = IDLE;
  static int address = 0;
  static uint16_t magic = 0;
  if(magic == 0){
    int seed = analogRead(10);
    randomSeed(seed);
    magic = random(65535);
  }

  unsigned long time = millis();
  buttonState = digitalRead(BUTTON_PIN);
  if(buttonState == HIGH && prevButtonState == LOW){
    Serial.println("Button down");
    pressTime = time;
  }
  else if(buttonState == LOW && prevButtonState == HIGH){
    Serial.println("Button up");
    if(time - pressTime > 1000){
      Serial.println("  Long press");
      Serial.println("  state -> DUMP");
      curMode = DUMP;
    }
    else{
      Serial.println("  Short press");
      if(curMode == IDLE){
        Serial.println("  state -> RECORD");
        address = 0;
        magic = 0;
	lastTime = 0;
        curMode = RECORD;
      }
      else if(curMode == RECORD){
        Serial.println("\n  state -> IDLE");
        curMode = IDLE;
      }
    }
  }
  if(curMode == DUMP){
    dump();
    curMode = IDLE;
  }
  else if(curMode == RECORD){
    if(lastTime == 0 || time/10 > lastTime/10){
      getGyroValues(&d);
      getAccelValues(&d);
      d.time = time;
      d.magic = magic;
      EEPROM1024.write_page(address, (uint8_t*)&d, 16);
      address += 16;
      if((address/16)%100 == 0)
        Serial.println(".");
      //printValues(&d);
      lastTime = time;
    }
  }
  prevButtonState = buttonState;
}

void dump(){
  struct imu_data d;
  uint8_t *b = (uint8_t*)&d;
  int i;
  int address = 0;
  uint16_t magic = 0;
  while(1){
    for(i=0; i<16; i++){
      b[i] = EEPROM1024.read(address + i);
    }
    if(address == 0)
      magic = d.magic;
    if(magic != d.magic)
      break;
    printValues(&d);
    address += 16;
  }
}

void printValues(struct imu_data *d){
  Serial.print(d->time);
  Serial.print(" ");
  Serial.print(d->gyro_x);
  Serial.print(" ");
  Serial.print(d->gyro_y);
  Serial.print(" ");
  Serial.print(d->gyro_z);
  Serial.print(" ");
  Serial.print(d->accel_x);
  Serial.print(" ");
  Serial.print(d->accel_y);
  Serial.print(" ");
  Serial.print(d->accel_z);
  Serial.println();
}

void setupLIS331DL(){
  //SPI 4/3 wire
  //1=ReBoot - reset chip defaults
  //n/a
  //filter off/on
  //filter for freefall 2
  //filter for freefall 1
  //filter freq MSB
  //filter freq LSB - Hipass filter (at 400hz) 00=8hz, 01=4hz, 10=2hz, 11=1hz (lower by 4x if sample rate is 100hz)
  writeRegister(LIS331DL_I2C_ADDR, LIS331DL_CTRL_REG2, 0b01000000);


  //sample rate 100/400hz
  //power off/on
  //2g/8g
  //self test
  //self test
  //z enable
  //y enable
  //x enable
  writeRegister(LIS331DL_I2C_ADDR, LIS331DL_CTRL_REG1, 0b01000111);
}

void getGyroValues(struct imu_data *d){
  byte xMSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_X_HIGH_REG);
  byte xLSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_X_LOW_REG);
  d->gyro_x = ((xMSB << 8) | xLSB);

  byte yMSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_Y_HIGH_REG);
  byte yLSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_Y_LOW_REG);
  d->gyro_y = ((yMSB << 8) | yLSB);

  byte zMSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_Z_HIGH_REG);
  byte zLSB = readRegister(L3G4200D_I2C_ADDR, L3G4200D_Z_LOW_REG);
  d->gyro_z = ((zMSB << 8) | zLSB);
}

void setupL3G4200D(int scale){
  //From  Jim Lindblom of Sparkfun's code
  // Enable x, y, z and turn off power down:
  writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG1, 0b00001111);

  // If you'd like to adjust/use the HPF, you can edit the line below to configure CTRL_REG2:
  writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG2, 0b00000000);

  // Configure CTRL_REG3 to generate data ready interrupt on INT2
  // No interrupts used on INT1, if you'd like to configure INT1
  // or INT2 otherwise, consult the datasheet:
  writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG3, 0b00001000);

  // CTRL_REG4 controls the full-scale range, among other things:

  if(scale == 250){
    writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG4, 0b00000000);
  }
  else if(scale == 500){
    writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG4, 0b00010000);
  }
  else{
    writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG4, 0b00110000);
  }

  // CTRL_REG5 controls high-pass filtering of outputs, use it
  // if you'd like:
  writeRegister(L3G4200D_I2C_ADDR, L3G4200D_CTRL_REG5, 0b00000000);
}

void getAccelValues(struct imu_data *d)
{
  //byte status = readRegister(LIS331DL_I2C_ADDR, LIS331DL_STATUS_REG);
  d->accel_x = readRegister(LIS331DL_I2C_ADDR, LIS331DL_X_HIGH_REG); 
  d->accel_y = readRegister(LIS331DL_I2C_ADDR, LIS331DL_Y_HIGH_REG); 
  d->accel_z = readRegister(LIS331DL_I2C_ADDR, LIS331DL_Z_HIGH_REG); 
}

//FIXME - needs error handling
void writeRegister(int deviceAddress, byte address, byte val) {
  Wire.beginTransmission(deviceAddress); // start transmission to device 
  Wire.write(address);                   // send register address
  Wire.write(val);                       // send value to write
  Wire.endTransmission();                // end transmission
}

//FIXME - needs error handling
int readRegister(int deviceAddress, byte address){
  int v;
  Wire.beginTransmission(deviceAddress);
  Wire.write(address); // register to read
  Wire.endTransmission();

  Wire.requestFrom(deviceAddress, 1); // read a byte

  //FIXME - needs timeout
  while(!Wire.available()) {
    // waiting
  }

  v = Wire.read();
  return v;
}
